#include	"defs.h"

typedef	struct {
	short	index;
	unsigned short	r;
	unsigned short	g;
	unsigned short	b;
} MAP8;

typedef struct {
	unsigned short	gb;
	short	next;
	unsigned short	r;
} MAP24;


static void
makepalette()
{
	int	i;

	for ( i = 0; i < MaxColorGuaranteed; i++ ){
		Palette[i] = ((Green[i] >> 3) & 0x1f) << 10
					| ((Red[i] >> 3) & 0x1f) << 5
					| ((Blue[i] >> 3) & 0x1f);
	}
}

static void
b2b( src, size )
void	*src;
int	size;
{
	unsigned short	*rptr= src;
	unsigned char	*wptr= src;

	size/= 2;

	while( --size >= 0 ){
		*wptr++= *rptr++;
	}
}

static void
b2w( src, size )
void	*src;
int	size;
{
	unsigned short	*rptr= src;
	unsigned short	*wptr= src;

	makepalette();
	size/= 2;

	while( --size >= 0 ){
		*wptr++= Palette[*rptr++];
	}
}

static void
w2b( src, size )
void	*src;
int	size;
{
	int	i,
		MaxColor= 0;
	short	*colormap, c;
	unsigned short	*rptr= src;
	unsigned char	*wptr= src;

	colormap= (short*)MALLOC( sizeof( short ) * 32768 );
	if( (int)colormap < 0 )
		error( "%s: Insufficient memory" );
	for( i= 0; i < 32768; i++ )	colormap[i]= -1;

	i= size / 2;
	while( --i >= 0 ){
		c= colormap[*rptr];
		if( c >= 0 ){
			*wptr++= c;
		}else {
			if( MaxColor >= 256 ){
				error ("%s: over 256 colors");
			}
			c= ( *rptr >> 2 ) & 0xf8;
			if( c & 8 )	c|= 7;
			Red[MaxColor]= c;
			c= ( *rptr >> 7 ) & 0xf8;
			if( c & 8 )	c|= 7;
			Green[MaxColor]= c;
			c= ( *rptr << 3 ) & 0xf8;
			if( c & 8 )	c|= 7;
			Blue[MaxColor]= c;
			colormap[*rptr]= MaxColor;
			*wptr++= MaxColor++;
		}
		rptr++;
	}

	MFREE( colormap );
	SETBLOCK( src, size / 2 );
	MaxColorGuaranteed= MaxColor;
}

static void
t2w( src, size )
void	*src;
int	size;
{
	unsigned char	*rptr= src;
	unsigned short	*wptr= src;
	unsigned char	r, g, b;

	size/= 2;

	while( --size >= 0 ){
		r= *rptr++ >> 3;
		g= *rptr++ >> 3;
		b= *rptr++ >> 3;

		*wptr++= ( g << 10 ) | ( r << 5 ) | b;
	}
	SETBLOCK( src, size );
}

static void
t2b( src, size )
void	*src;
int	size;
{
	int	i;
	int	MaxColor= 0;
	unsigned char	*rptr= (unsigned char*)src;
	unsigned char	*wptr= (unsigned char*)src;
	unsigned short	r, gb;
	short	c, colormap[256];
	MAP24	palet[256];

	for( i= 0; i < 256; i++ ){
		colormap[i]= -1;
		palet[i].next= -1;
	}

	i= size / 2;
	while( --i >= 0 ){
		r= *rptr++;
		gb= *rptr++ << 8;
		gb|= *rptr++;
		c= colormap[r];
		if( c < 0 ){
			if( MaxColor >= 256 ){
				error ("%s: over 256 colors");
			}
			colormap[r]= MaxColor;
			palet[MaxColor].gb= gb;
			palet[MaxColor].r= r;
			*wptr++= MaxColor++;
		}else {
			while( 1 ){
				if( palet[c].gb == gb ){
					*wptr++= c;
					break;
				}
				if( palet[c].next < 0 ){
					if( MaxColor >= 256 ){
						error ("%s: over 256 colors");
					}
					palet[c].next= MaxColor;
					palet[MaxColor].gb= gb;
					palet[MaxColor].r= r;
					*wptr++= MaxColor++;
					break;
				}
				c= palet[c].next;
			}
		}
	}
	SETBLOCK( src, size / 2 );

	for( i= 0; i < MaxColor; i++ ){
		Red[i]= palet[i].r;
		Green[i]= ( palet[i].gb >> 8 ) & 0xff;
		Blue[i]= palet[i].gb & 0xff;
	}
	MaxColorGuaranteed= MaxColor;
}

static void
cc8( src, size )
void	*src;
int	size;
{
	int	i,
		MaxColor= 0;
	short	c;
	MAP8	colormap[256];
	unsigned short	*rptr= src;
	unsigned char	*wptr;

	for( i= 0; i < 256; i++ )	colormap[i].index= -1;

	InBpp= 1;

	i= size / 2;
	while( --i >= 0 ){
		c= *rptr;
		if( colormap[c].index < 0 ){
			colormap[c].index= MaxColor;
			colormap[MaxColor].r= Red[c];
			colormap[MaxColor].g= Green[c];
			colormap[MaxColor++].b= Blue[c];
		}
		rptr++;
	}

	if( MaxColorGuaranteed == MaxColor ){
		b2b( src, size );
	}else {
		rptr= src;
		wptr= src;
		i= size / 2;
		while( --i >= 0 ){
			*wptr++= colormap[*rptr++].index;
		}

		for( i= 0; i < 256; i++ ){
			c= colormap[i].index;
			if( c >= 0 ){
				Red[c]= colormap[i].r;
				Green[c]= colormap[i].g;
				Blue[c]= colormap[i].b;
			}
		}
		MaxColorGuaranteed= MaxColor;
	}
}

static void
cc15( src, size )
void	*src;
int	size;
{
	int	i,
		MaxColor= 0;
	short	*colormap, c;
	unsigned short	*rptr= src;
	unsigned char	*wptr;

	colormap= (short*)MALLOC( sizeof( short ) * 32768 );
	if( (int)colormap < 0 )
		error( "%s: Insufficient memory" );
	for( i= 0; i < 32768; i++ )	colormap[i]= -1;

	i= size / 2;
	while( --i >= 0 ){
		c= colormap[*rptr];
		if( c < 0 ){
			if( MaxColor >= 256 ){
				MFREE( colormap );
				return;
			}
			c= ( *rptr >> 2 ) & 0xf8;
			if( c & 8 )	c|= 7;
			Red[MaxColor]= c;
			c= ( *rptr >> 7 ) & 0xf8;
			if( c & 8 )	c|= 7;
			Green[MaxColor]= c;
			c= ( *rptr << 3 ) & 0xf8;
			if( c & 8 )	c|= 7;
			Blue[MaxColor]= c;
			colormap[*rptr]= MaxColor++;
		}
		rptr++;
	}

	rptr= src;
	wptr= src;
	i= size / 2;
	while( --i >= 0 ){
		*wptr++= colormap[*rptr++];
	}

	MFREE( colormap );
	SETBLOCK( src, size / 2 );
	MaxColorGuaranteed= MaxColor;
	InBpp= 1;
}

static void
cc24( src, size )
void	*src;
int	size;
{
	int	i;
	int	MaxColor= 0;
	unsigned char	*rptr= (unsigned char*)src;
	unsigned char	*wptr;
	unsigned short	r, gb;
	short	c, colormap[256];
	MAP24	palet[256];

	for( i= 0; i < 256; i++ ){
		colormap[i]= -1;
		palet[i].next= -1;
	}

	i= size / 2;
	while( --i >= 0 ){
		r= *rptr++;
		gb= *rptr++ << 8;
		gb|= *rptr++;
		c= colormap[r];
		if( c < 0 ){
			if( MaxColor >= 256 ){
				MaxColorGuaranteed= 16777216;
				InBpp= 4;
				return;
			}
			colormap[r]= MaxColor;
			palet[MaxColor].gb= gb;
			palet[MaxColor++].r= r;
		}else {
			while( 1 ){
				if( palet[c].gb == gb ){
					break;
				}
				if( palet[c].next < 0 ){
					if( MaxColor >= 256 ){
						MaxColorGuaranteed= 16777216;
						InBpp= 4;
						return;
					}
					palet[c].next= MaxColor;
					palet[MaxColor].gb= gb;
					palet[MaxColor++].r= r;
					break;
				}
				c= palet[c].next;
			}
		}
	}

	rptr= src;
	wptr= src;
	i= size / 2;
	while( --i >= 0 ){
		r= *rptr++;
		gb= *rptr++ << 8;
		gb|= *rptr++;
		c= colormap[r];
		while( palet[c].gb != gb ){
			c= palet[c].next;
		}
		*wptr++= c;
	}
	SETBLOCK( src, size / 2 );

	for( i= 0; i < MaxColor; i++ ){
		Red[i]= palet[i].r;
		Green[i]= ( palet[i].gb >> 8 ) & 0xff;
		Blue[i]= palet[i].gb & 0xff;
	}
	MaxColorGuaranteed= MaxColor;
	InBpp= 1;
}

void
ColorManager()
{
	if( OutBpp == 1 ){
		if( InBpp == 2 ){
			w2b( Screen, Size );
		}else if( InBpp & 4 ){
			t2b( Screen, Size );
		}else if( !( InBpp & 0x40 ) ){
			b2b( Screen, Size );
		}
		InBpp= 1;
	}else if( OutBpp == 2 ){
		if( InBpp == 1 ){
			b2w( Screen, Size );
		}else if( InBpp & 4 ){
			t2w( Screen, Size );
		}
		InBpp= 2;
	}else if( OutBpp & 0x80 ){
		if( InBpp == 1 ){
			cc8( Screen, Size );
		}else if( InBpp == 2 ){
			cc15( Screen, Size );
		}else {
			cc24( Screen, Size );
		}
	}
}
